/*
 *			SPC_SCSI.C for Minix-vmd
 *
 * This is the driver for SCSI adapters using FUJITSU MB89352A aka SPC,
 * based on beta 4 driver for minix 2.0.0.  The original SCSI driver for
 * Adaptec 154X SCSI Host Adapter family was written by James da Silva
 * (jds@cs.umd.edu) and has been extensively revised by Kees J. Bot
 * (kjb@cs.vu.nl).
 *
 *			Beta 1	March 30, 1997		Osamu Ishibashi
 *			Beta 2	November 5, 1997	Osamu Ishibashi
 *			Beta 3	November 7, 1997	Osamu Ishibashi
 *			Beta 4	November 28, 1997	Osamu Ishibashi
 *			Minix-vmd December 5, 1997	Osamu Ishibashi
 *
 * The file contains one entry point:
 *
 *   spc_scsi_task:	main entry when system is brought up
 *
 */
#include "kernel.h"
#include "driver.h"
#include "drvlib.h"
#if ENABLE_SPC_SCSI
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>
#include "scsi.h"
#include "assert.h"
INIT_ASSERT


#ifndef SCSI_DEBUG
#define SCSI_DEBUG	0	/* 1=dump SCSI bus data | 2=dump regs
				 * 4=show request | 8=dump scsi cmd
				 */
#endif

#define HOST_ID		7

/* SPC port addresses */
#define SPC_BASEREG	0x300	/* default base port address */
#define SPC_BDID	spc_base		/* controller ID */
#define SPC_CTL		(spc_base+spc_offs)	/* define mode of actions */
#define SPC_CMD		(spc_base+2*spc_offs)	/* command register */
#define SPC_INTS	(spc_base+4*spc_offs)	/* sense/reset int */
#define SPC_PSNS	(spc_base+5*spc_offs)	/* sense SCSI control lines */
#define SPC_DGC		(spc_base+5*spc_offs)	/* diagnostic mode control */
#define SPC_STS		(spc_base+6*spc_offs)	/* controller status */
#define SPC_ERR		(spc_base+7*spc_offs)	/* controller error status */
#define SPC_PCTL	(spc_base+8*spc_offs)	/* instruct SCSI phase */
#define SPC_MBC		(spc_base+9*spc_offs)	/* modified byte counter */
#define SPC_DREG	(spc_base+10*spc_offs)	/* SCSI transfer 8-byte FIFO */
#define SPC_TEMP	(spc_base+11*spc_offs)	/* SCSI bus data register */
#define SPC_TCH		(spc_base+12*spc_offs)	/* transfer counter high byte */
#define SPC_TCM		(spc_base+13*spc_offs)	/* transfer counter mid byte */
#define SPC_TCL		(spc_base+14*spc_offs)	/* transfer counter low byte */

/* control register (SCTL) bits */
#define SPC_RESET	0x80	/* reset the controller */
#define SPC_ARBIT	0x10	/* enable arbitration upon selection */
#define SPC_PARITY	0x08	/* enable parity check */
#define SPC_INT		0x01	/* enable interrupt */

/* SPC command codes (bit 5 - 7 of SCMD register) */
#define SPC_BUS_RELEASE	0x00	/* bus release command */
#define SPC_SELECT	0x20	/* select command */
#define SPC_RESET_ATN	0x40	/* reset ATN command */
#define SPC_SET_ATN	0x60	/* set ATN command */
#define SPC_TRANSFER	0x80	/* transfer command */
#define SPC_PAUSE	0xA0	/* transfer Pause command */
#define SPC_RESET_AR	0xC0	/* reset ACK/REQ command */
#define SPC_SET_AR	0xE0	/* set ACK/REQ command */
#define SPC_RST_OUT	0x10	/* output SCSI reset signal */

/* SCMD low three bits modifies transfer mode */
#define SPC_DMA		0x00	/* perform DMA transfer */
#define SPC_PIO		0x04	/* perform PIO transfer */
#define SPC_TRM_MODE	0x01	/* abort transfer on a parity error */

/* phase sense register (PSNS) bits */
#define SCSI_REQ	0x80	/* SCSI bus REQ */
#define SCSI_ACK	0x40	/* SCSI bus ACK */
#define SCSI_ATN	0x20	/* SCSI bus ATN */
#define SCSI_SEL	0x10	/* SCSI bus SEL */
#define SCSI_BSY	0x08	/* SCSI bus BSY */
#define SCSI_MSG	0x04	/* SCSI bus MSG */
#define SCSI_CD		0x02	/* SCSI bus C/D */
#define SCSI_IO		0x01	/* SCSI bus IO */
#define PHASE_MASK	7
#define bus_phase()	(in_byte(SPC_PSNS) & PHASE_MASK)

/* Diagnostic mode control (SDGC) register */
#define SPC_XFER_EA	0x20	/* enable int on PIO transfer;
				   this bit actually has nothing to do w/
				   diagnostic mode */
#define SPC_XFER_DA	0

/* int sense register (INTS) bits */
#define SPC_DISCOND	0x20	/* target has disconnected */
#define SPC_CMD_CMPL	0x10	/* select command done */
#define SPC_SVC_REQ	0x08	/* service required */
#define SPC_TIMEOUT	0x04	/* selection timeout */
#define SPC_RESET_CND	0x01	/* SCSI bus reset detected */

/* SSTS high two bits indicates connection status */
#define SPC_INIT	0x80	/* connected as initiater */
#define SPC_TARGET	0x40	/* connected as target */
#define SPC_FREE	0	/* no connection */
#define spc_phase()	(in_byte(SPC_STS) & 0xC0)

/* SSTS other interesting bits */
#define SPC_TC_ZERO	0x04	/* TC registers are 0 */
#define SPC_DREG_FULL	0x02	/* FIFO to read has got full */
#define SPC_DREG_EMPTY	0x01	/* FIFO to write has got empty */

/* phase control register (PCTL) bits */
#define SPC_BFI_EA	0x80	/* enable bus free interrupt */
#define SPC_BFI_DA	0x00	/* disable bus free interrupt */

/* PCTL low 3 bits are common with PSNS */
#define SCSI_DATA_OUT	0x00	/* data out phase */
#define SCSI_DATA_IN	0x01	/* data in phase */
#define SCSI_COMMAND	0x02	/* command phase */
#define SCSI_STATUS	0x03	/* status phase */
#define SCSI_MSG_OUT	0x06	/* message out phase */
#define SCSI_MSG_IN	0x07	/* message in phase */

#define SPC_FIFO_SIZE	   8	/* FIFO buffer size */


/* Command Control Block structure */
typedef struct {
    byte cmdlen;		/* SCSI Command Length (6 for Group 0) */
    byte senselen;		/* Request/Disable Sense, Allocation Length */
    byte cmd[sizeof(cdb1_t)];	/* SCSI Command Descriptor Block */
    byte sense[sizeof(sense_t)];/* SCSI Request Sense Information */
} ccb_t;

/* DMA request struct */
typedef struct {
    phys_bytes datalen;		/* number of bytes to transfer */
    phys_bytes dataptr;		/* address of transfer memory */
} dma_t;

/* Administration for one SCSI request. */
typedef struct request {
  int opcode;			/* DEV_GATHER or DEV_SCATTER */
  unsigned count;		/* number of bytes to transfer */
  unsigned remain;		/* number of bytes not yet transfered */
  unsigned retry;		/* number of tries allowed if retryable */
  u64_t position;		/* first byte on the device to transfer */
  ccb_t ccb;			/* Command Control Block */
  dma_t dmalist[NR_IOREQS];	/* scatter/gather dma list */
  dma_t *dmaptr;		/* pointer to dmalist[] */
  struct iorequest_s *iov[NR_IOREQS];	/* affected I/O requests */
} request_t;


/* Variables */

PRIVATE request_t request;	/* current request (there is only one) */
	#define rq		(&request)
	#define ccb_cmd0(rq)	(* (cdb0_t *) (rq)->ccb.cmd)
	#define ccb_cmd1(rq)	(* (cdb1_t *) (rq)->ccb.cmd)
	#define ccb_sense(rq)	(* (sense_t *) (rq)->ccb.sense)
	#define dptr		(request.dmaptr->dataptr)
	#define dlen		(request.dmaptr->datalen)
PRIVATE int spc_tasknr = ANY;	/* Any to prevent accidents */
PRIVATE int spc_base;		/* base I/O register */
PRIVATE int spc_offs;		/* I/O register offset */
PRIVATE int spc_status;		/* controller probe result */
PRIVATE struct scsi *s_sp;	/* active SCSI device struct */
PRIVATE struct device *s_dv;	/* active partition */
PRIVATE int s_type;		/* sd, rst, nrst? */
PRIVATE unsigned long s_nextpos;/* next byte on the device to transfer */
PRIVATE unsigned long s_buf_blk;/* disk block currently in tmp_buf */
PRIVATE int s_must;		/* must finish the current request? */
PRIVATE int spc_irq;		/* configured IRQ */
PRIVATE int spc_dma;		/* configured DMA */
PRIVATE inquiry_t inqdata;	/* results of Inquiry command */
PRIVATE message mess;		/* to receive hardware interrupt */
PRIVATE void (*port_rw_byte)(unsigned port, phys_bytes ptr, unsigned count);
				/* pointer to port read/write */
PRIVATE int ready_bit;		/* mask pattern to block wait while transfer */
PRIVATE char *phase[8] = {	/* to print out SCSI phase on fatal error */
	"data out", "data in", "command", "status",
	"impossible", "impossible", "message out", "message in"
	};

/* Functions */

FORWARD _PROTOTYPE( struct device *s_prepare, (int device) );
FORWARD _PROTOTYPE( char *s_name, (void) );
FORWARD _PROTOTYPE( int s_do_open, (struct driver *dp, message *m_ptr) );
FORWARD _PROTOTYPE( int scsi_probe, (void) );
FORWARD _PROTOTYPE( int scsi_inquiry, (void) );
FORWARD _PROTOTYPE( int scsi_ndisk, (void) );
FORWARD _PROTOTYPE( int scsi_ntape, (void) );
FORWARD _PROTOTYPE( int s_transfer, (int proc_nr, int opcode, u64_t position,
					iovec_t *iov, unsigned nr_req) );
FORWARD _PROTOTYPE( int s_rdcdrom, (int proc_nr, int opcode, u64_t position,
					iovec_t *iov, unsigned nr_req) );
FORWARD _PROTOTYPE( int s_do_close, (struct driver *dp, message *m_ptr) );
FORWARD _PROTOTYPE( int s_do_ioctl, (struct driver *dp, message *m_ptr) );
FORWARD _PROTOTYPE( void s_geometry, (struct partition *entry));
FORWARD _PROTOTYPE( int scsi_simple, (int opcode, int count) );
FORWARD _PROTOTYPE( void group0, (void) );
FORWARD _PROTOTYPE( void group1, (void) );
FORWARD _PROTOTYPE( int scsi_command, (byte *data, int len) );
FORWARD _PROTOTYPE( void spc_pio, (void) );
FORWARD _PROTOTYPE( void scsi_in, (int phase, byte *data, int len) );
FORWARD _PROTOTYPE( void scsi_out, (int phase, byte *data, int len) );
FORWARD _PROTOTYPE( void spc_init, (void) );
FORWARD _PROTOTYPE( int s_handler, (int irq) );

FORWARD _PROTOTYPE( void h2b16, (big16 b, U16_t l) );
FORWARD _PROTOTYPE( void h2b24, (big24 b, u32_t l) );
FORWARD _PROTOTYPE( void h2b32, (big32 b, u32_t l) );
FORWARD _PROTOTYPE( u16_t b2h16, (big16 b) );
FORWARD _PROTOTYPE( u32_t b2h24, (big24 b) );
FORWARD _PROTOTYPE( u32_t b2h32, (big32 b) );

#define spc_tc()	((long)in_byte(SPC_TCH) << 16 \
			| in_byte(SPC_TCM) << 8 \
			| in_byte(SPC_TCL))
#define wait_int()	{if (spc_irq < 0) \
			{while(!in_byte(SPC_INTS)); out_byte(SPC_INTS, 0xFF);} \
			else receive(HARDWARE, &mess);}


/* Some of the above errors must be printed on the console. */
#if SCSI_DEBUG & 1
#define sense_serious(key)	((key) != 0)
#else
#define sense_serious(key)	((0xFE1C & (1 << (key))) != 0)
#endif

#if SCSI_DEBUG & 2
#define dump_regs(s)	printf \
			("%s: INTS=%x, PSNS=%x, SSTS=%x, PCTL=%x, TC=%lu\n", \
			(s), in_byte(SPC_INTS), in_byte(SPC_PSNS), \
			in_byte(SPC_STS), in_byte(SPC_PCTL), spc_tc())
#else
#define dump_regs(s)
#endif

#if SCSI_DEBUG & 4
FORWARD _PROTOTYPE( void show_iov, (iovec_t *iov, unsigned nr_req) );
FORWARD _PROTOTYPE( void show_req, (void) );
#else
#define show_iov(iov, nr_req)
#define show_req()
#endif

#if SCSI_DEBUG & 8
FORWARD _PROTOTYPE( void dump_scsi_cmd, (void) );
#else
#define dump_scsi_cmd()
#endif


/* Entry points to this driver. */
PRIVATE struct driver s_dtab = {
  s_name,	/* current device's name */
  s_do_open,	/* open or mount request, initialize device */
  s_do_close,	/* release device */
  s_do_ioctl,	/* tape and partition ioctls */
  s_prepare,	/* prepare for I/O on a given minor device */
  s_transfer,	/* do the I/O */
  nop_cleanup,	/* no cleanup needed */
  s_geometry	/* tell the geometry of the disk */
};


/*===========================================================================*
 *				spc_scsi_task				     *
 *===========================================================================*/
PUBLIC void spc_scsi_task()
{
/* Set target and logical unit numbers, then call the generic main loop. */
  int i;
  struct scsi *sp;
  long v;
  char *name;
  static char fmt[] = "d,d";

  spc_tasknr = proc_number(proc_ptr);

  for (i = 0; i < MAX_DEVICES; i++) {
	(void) s_prepare(i * DEV_PER_DRIVE);
	sp = s_sp;

	/* Look into the environment for special parameters. */
	name = s_name();

	v = i;
	(void) env_parse(name, fmt, 0, &v, 0L, 7L);
	sp->targ = v;

	v = 0;
	(void) env_parse(name, fmt, 1, &v, 0L, 7L);
	sp->lun = v;
  }

  spc_init();

  driver_task(&s_dtab);
}


/*===========================================================================*
 *				s_prepare				     *
 *===========================================================================*/
PRIVATE struct device *s_prepare(device)
int device;
{
/* Prepare for I/O on a device. */

  rq->count = 0;	/* no requests as yet */
  s_must = TRUE;	/* the first transfers must be done */
  s_buf_blk = -1;	/* invalidate s_buf_blk */

  if (device < NR_DISKDEVS) {			/* sd0, sd1, ... */
	s_type = TYPE_SD;
	s_sp = &scsi[device / DEV_PER_DRIVE];
	s_dv = &s_sp->part[device % DEV_PER_DRIVE];
  } else
  if ((unsigned) (device - MINOR_hd1a) < NR_SUBDEVS) {	/* sd1a, sd1b, ... */
	device -= MINOR_hd1a;
	s_type = TYPE_SD;
	s_sp = &scsi[device / SUB_PER_DRIVE];
	s_dv = &s_sp->subpart[device % SUB_PER_DRIVE];
  } else
  if ((unsigned) (device - MINOR_st0) < NR_TAPEDEVS) {	/* nrst0, rst0, ... */
	device -= MINOR_st0;
	s_type = device & 1 ? TYPE_RST : TYPE_NRST;
	s_sp = &scsi[device >> 1];
	s_dv = &s_sp->dummypart;
  } else {
	return(NIL_DEV);
  }

  return(s_dv);
}


/*===========================================================================*
 *				s_name					     *
 *===========================================================================*/
PRIVATE char *s_name()
{
/* Return a name for the current device. */
  static char name[] = "sd35";
  int n = (s_sp - scsi);

  switch (s_type) {
  case TYPE_SD:			/* Disk device: sd* */
	name[1] = 'd';
	n *= DEV_PER_DRIVE;
	break;
  case TYPE_RST:		/* Tape device: st* */
  case TYPE_NRST:
	name[1] = 't';
	break;
  }
  if (n < 10) {
	name[2] = '0' + n;
	name[3] = 0;
  } else {
	name[2] = '0' + n / 10;
	name[3] = '0' + n % 10;
  }
  return name;
}


/*===========================================================================*
 *				s_do_open				     *
 *===========================================================================*/
PRIVATE int s_do_open(dp, m_ptr)
struct driver *dp;
message *m_ptr;
{
  struct scsi *sp;
  int r;

  if (spc_status != OK) return(EIO); /* no controller, forget it */

  if (s_prepare(m_ptr->DEVICE) == NIL_DEV) return(ENXIO);
  sp = s_sp;

  if ((r = scsi_probe()) != OK) return(r);

  if (sp->state & S_RDONLY && m_ptr->COUNT & W_BIT) {
	printf("%s: write-protected\n", s_name());
	return(EACCES);
  }

  switch (sp->devtype) {
  case SCSI_DEVDISK:
  case SCSI_DEVWORM:
  case SCSI_DEVCDROM:
  case SCSI_DEVOPTICAL:
	/* Lock the medium and read partition tables on first open. */
	if (sp->open_ct == 0) {
		(void) scsi_simple(SCSI_REMOVAL, 1);
		partition(&s_dtab, (int) (sp-scsi) * DEV_PER_DRIVE, P_PRIMARY);
	}
	break;
  case SCSI_DEVTAPE:
	/* Make sure tape is not already open. */
	if (sp->open_ct > 0) return(EBUSY);

	sp->open_mode = m_ptr->COUNT;
	/* If open(..., O_WRONLY) then write a filemark on close even if no
	 * write is done.
	 */
	sp->need_eof = ((sp->open_mode & (R_BIT|W_BIT)) == W_BIT);
	break;
  }
  sp->open_ct++;
  return(OK);
}


/*===========================================================================*
 *				scsi_probe				     *
 *===========================================================================*/
PRIVATE int scsi_probe()
{
/* See if a device exists and if it is ready. */
  struct scsi *sp = s_sp;
  sense_t *sense;
  int r, key;

  /* Something out there? */
  if (in_byte(SPC_PSNS) == 0xFF)
	return(ENODEV);	/* nothing's hooked up */

  r = scsi_simple(SCSI_REQSENSE, sizeof(sense_t));
  switch (r) {
  case SENSE_NO_SENSE:
	break;
  case SENSE_NOT_READY:
  case SENSE_UNIT_ATT:
	/* Device is in a "look at me" state, probably changed media. */
	sp->state &= ~S_READY;
	break;
  case SENSE_ILLEGAL_REQ:
	printf("%s: offline\n", s_name());
	/* fall through */
  case SENSE_HARDWARE:
	return(EIO);
  default:
	printf("request sense code %x for %s\n", r, s_name());
	return(EIO);
  }

  if (!(sp->state & S_PRESENT)) {
	/* First contact with a new device, what type is it? */

	if ((r = scsi_inquiry()) != OK) return(r);

	sp->devtype = inqdata.devtype;
  }

  if (!(sp->state & S_READY)) {
	/* If it's a disk: start it, if it's a tape: load it. */
	(void) scsi_simple(SCSI_STRTSTP, 1);
  }

  /* See if the unit is ready for I/O.  A disk may be spinning up, a
   * floppy or tape drive may be empty.
   */
  while ((key = scsi_simple(SCSI_UNITRDY, 0)) != SENSE_NO_SENSE) {
	/* Not ready, why? */

	sp->state &= ~S_READY;

	switch (key) {
	case SENSE_UNIT_ATT:
		/* A media change or something, try again. */
		break;
	case SENSE_NOT_READY:
		/* Look at the additional sense data to see why it isn't
		 * ready.
		 */
		sense = &ccb_sense(rq);
		switch ((sense->add_code << 8) | sense->add_qual) {
		case 0x0401:
			/* "It is becoming ready."  Fine, we wait. */
			milli_delay(1000);
			break;
		case 0x0402:
			/* "Initialization command required."  So we tell it
			 * to spin up.
			 */
			if (scsi_simple(SCSI_STRTSTP, 1) != SENSE_NO_SENSE)
				return(EIO);
			break;
		case 0x0403:
			/* "Manual intervention required." */
		case 0x3A00:
			/* "No media present." */
			printf("%s: no media loaded\n", s_name());
			return(EIO);
		default:
			/* For some reason it is not usable. */
			printf("%s: not ready\n", s_name());
			return(EIO);
		}
		break;
	default:
		/* The device is in some odd state.  */
		if (key != SENSE_NOT_READY) {
			printf("%s: hardware error\n", s_name());
			return(EIO);
		}
	}
  }

  if (!(sp->state & S_PRESENT)) {
	/* Do the inquiry again, the message may have changed. */
	if (scsi_inquiry() != OK) return(EIO);

	/* Tell what kind of device it is we have found. */

	printf("%s: %-7s %.48s\n",
		s_name(),
		inqdata.devtype > SCSI_DEVMAX ? "UNKNOWN"
				: scsi_devstr[inqdata.devtype],
		inqdata.vendor /* + product + revision + extra */);
  }

  if (!(sp->state & S_READY)) {
	/* Get the geometry, limits, etc. */

	switch (sp->devtype) {
	case SCSI_DEVDISK:
	case SCSI_DEVWORM:
	case SCSI_DEVCDROM:
	case SCSI_DEVOPTICAL:
		if (scsi_ndisk() != OK) return(EIO);
		break;
	case SCSI_DEVTAPE:
		if (scsi_ntape() != OK) return(EIO);
		break;
	default:
		printf("%s: unsupported\n", s_name());
		return(EIO);
	}
  }
  return(OK);
}


/*===========================================================================*
 *				scsi_inquiry				     *
 *===========================================================================*/
PRIVATE int scsi_inquiry()
{
  /* Prefill with nulls. */
  memset(tmp_buf, '\0', sizeof(inquiry_t));

  /* Do a SCSI inquiry. */
  if (scsi_simple(SCSI_INQUIRY, sizeof(inquiry_t)) != SENSE_NO_SENSE)
	return(EIO);
  inqdata = * (inquiry_t *) tmp_buf;

  if (inqdata.len == 0) {
	/* The device doesn't return meaningful text fields. */
	strcpy(inqdata.vendor, "(unknown)");
  }

  /* The top three bits of devtype must be zero for the lun to exist. */
  if ((inqdata.devtype & 0xE0) != 0) return(ENXIO);

  return(OK);
}


/*===========================================================================*
 *				scsi_ndisk				     *
 *===========================================================================*/
PRIVATE int scsi_ndisk()
{
/* Gather disk data, capacity and block size. */

  struct scsi *sp = s_sp;
  unsigned long capacity = -1, block_size = SECTOR_SIZE;
  byte *buf = tmp_buf;

  /* Minor device type must be for a disk. */
  if (s_type != TYPE_SD) return(EIO);

  if (sp->devtype == SCSI_DEVCDROM) {
	/* Read-only by definition. */
	sp->state |= S_RDONLY;
  } else {
	/* SCSI modesense to find out if the disk is write protected. */
	if (scsi_simple(SCSI_MDSENSE, 255) != SENSE_NO_SENSE) return(EIO);

	/* Write protected? */
	sp->state &= ~S_RDONLY;
	if (buf[2] & 0x80) sp->state |= S_RDONLY;

	/* Don't write a worm disk, not wise at the moment. */
	if (sp->devtype == SCSI_DEVWORM) sp->state |= S_RDONLY;
  }

  /* Get drive capacity and block size. */
  group1();
  ccb_cmd1(rq).scsi_op = SCSI_CAPACITY;

  if (scsi_command(tmp_buf, 8) == SENSE_NO_SENSE) {
	capacity = b2h32(buf + 0) + 1;
	block_size = b2h32(buf + 4);
	printf("%s: capacity %lu x %lu bytes\n",
					s_name(), capacity, block_size);
  } else {
	printf("%s: unknown capacity\n", s_name());
  }

  /* We do not believe block sizes over MAX_BLOCK_SIZE. */
  if (block_size > MAX_BLOCK_SIZE) {
	printf("%s: can't handle %lu byte blocks\n", s_name(), block_size);
	return(EIO);
  }

  sp->block_size = block_size;
#if _WORD_SIZE > 2
  /* Keep it within reach of a group 0 command. */
  sp->count_max = 0x100 * block_size;
#else
  sp->count_max = block_size > UINT_MAX/0x100 ? UINT_MAX : 0x100 * block_size;
#endif

  /* The fun ends at 4GB. */
  if (capacity > ((unsigned long) -1) / block_size)
	sp->part[0].dv_size = mul64u(-1, block_size);
  else
	sp->part[0].dv_size = mul64u(capacity, block_size);

  /* Finally we recognize its existence. */
  sp->state |= S_PRESENT|S_READY;

  return(OK);
}


/*===========================================================================*
 *				scsi_ntape				     *
 *===========================================================================*/
PRIVATE int scsi_ntape()
{
/* Gather tape data, block limits, fixed block size or not. */
  struct scsi *sp = s_sp;
  unsigned minblk;
  unsigned long maxblk;
  byte *buf = tmp_buf;

  /* Minor device type must be for a tape. */
  if (s_type != TYPE_RST && s_type != TYPE_NRST) return(EIO);

  /* Read limits. */
  if (scsi_simple(SCSI_RDLIMITS, 6) != SENSE_NO_SENSE) return(EIO);
  minblk = b2h16(buf + 4);
  maxblk = b2h24(buf + 1);

  printf("%s: limits: min block len %u, max block len %lu\n",
	s_name(), minblk, maxblk);

  if (sp->state & S_PRESENT) {
	/* Keep the current block size. */
	if (sp->tfixed) minblk= maxblk= sp->block_size;
  }

  sp->tstat.mt_dsreg = DS_OK;
  sp->tstat.mt_erreg = 0;
  sp->tstat.mt_fileno = 0;
  sp->tstat.mt_blkno = 0;
  sp->tstat.mt_resid = 0;

  if (minblk == maxblk) {
	/* Fixed block length. */
	sp->tfixed = TRUE;
	sp->block_size = minblk;
	sp->tstat.mt_blksize = minblk;
	sp->count_max = UINT_MAX;
  } else {
	/* Variable block length. */
	sp->tfixed = FALSE;
	sp->block_size = 1;
	sp->tstat.mt_blksize = 0;
	sp->count_max = maxblk == 0 ? UINT_MAX : maxblk;
  }

  /* SCSI modesense. */
  if (scsi_simple(SCSI_MDSENSE, 255) != SENSE_NO_SENSE) return(EIO);

  /* Write protected? */
  sp->state &= ~S_RDONLY;
  if (buf[2] & 0x80) sp->state |= S_RDONLY;

  /* Density and block size. */
  if (buf[3] >= 8) {
	printf("%s: density 0x%02x, nblocks %lu, block len ",
		s_name(),
		buf[4],
		b2h24(buf + 4 + 1));
	printf(sp->tfixed ? "%lu\n" : "variable\n", b2h24(buf + 4 + 5));
  }

  sp->state |= S_PRESENT|S_READY;
  return(OK);
}


/*===========================================================================*
 *				s_transfer				     *
 *===========================================================================*/
PRIVATE int s_transfer(proc_nr, opcode, position, iov, nr_req)
int proc_nr;			/* process doing the request */
int opcode;			/* DEV_GATHER or DEV_SCATTER */
u64_t position;			/* offset on device to read or write */
iovec_t *iov;			/* pointer to read or write request vector */
unsigned nr_req;		/* length of request vector */
{
  struct scsi *sp = s_sp;
  iovec_t *iop, *iov_end = iov + nr_req;
  int key;
  unsigned nbytes, dma_count, chunk;
  vir_bytes addr, size;
  phys_bytes dma_phys;
  phys_bytes last_phys;
  unsigned last_count;
  unsigned long block;
  phys_bytes user_base = proc_vir2phys(proc_addr(proc_nr), 0);

  rq->opcode = opcode;
  rq->retry = 2;

  while (nr_req > 0) {
	/* Probe the device if it isn't ready. */
	if (!(sp->state & S_READY) && scsi_probe() != OK) return(EIO);

	show_iov(iov, nr_req);

	/* How many bytes to transfer? */
	nbytes = 0;
	for (iop = iov; iop < iov_end; iop++) nbytes += iop->iov_size;

	switch (sp->devtype) {
	case SCSI_DEVCDROM:
	case SCSI_DEVWORM:
	case SCSI_DEVDISK:
	case SCSI_DEVOPTICAL:
		/* Which block on disk and how close to EOF? */
		if (cmp64(position, s_dv->dv_size) >= 0) return(OK);  /* EOF */
		if (cmp64(add64u(position, nbytes), s_dv->dv_size) > 0)
			nbytes = diff64(s_dv->dv_size, position);
		rq->position = add64(s_dv->dv_base, position);

		if ((nbytes % sp->block_size) != 0
				|| rem64u(rq->position, sp->block_size) != 0) {
			/* Not on a device block boundary.  CD-ROM? */
			return(s_rdcdrom(proc_nr, opcode, position, iov,
								nr_req));
		}
		break;

	case SCSI_DEVTAPE:
		if ((nbytes % sp->block_size) != 0) return(EINVAL);

		/* Old error condition? */
		if (sp->tstat.mt_dsreg == DS_ERR) return(EIO);

		if (opcode == DEV_GATHER && sp->at_eof) return(OK);
		break;

	default:
		return(EIO);
	}

	/* Construct a DMA scatter/gather vector from the I/O vector. */
	rq->count = 0;
	rq->dmaptr = rq->dmalist;
	for (iop = iov; iop < iov_end; iop++) {
		addr = iop->iov_addr;
		size = iop->iov_size;

		while (size > 0) {
			/* Page translation. */
			dma_phys = phys2bus(user_base + addr);
			dma_count = page_bytes_left(dma_phys);

			/* Several constraints... */
			if (dma_count > size)
				dma_count = size;
			if (rq->count + dma_count > nbytes)
				dma_count = nbytes - rq->count;
			if (rq->count + dma_count > sp->count_max)
				dma_count = sp->count_max - rq->count;

			if (rq->count > 0 && dma_phys == last_phys) {
				/* Two blocks are adjacent! */
				rq->dmaptr[-1].datalen =
					(last_count += dma_count);
			} else
			if (rq->dmaptr < &rq->dmalist[NR_IOREQS]
				&& (rq->count == 0
					|| DMA_CHECK(last_phys, dma_phys))
			) {
				/* New scatter/gather entry. */
				rq->dmaptr->dataptr = dma_phys;
				rq->dmaptr->datalen = last_count = dma_count;
				rq->dmaptr++;
			} else {
				/* End of DMA vector, or addresses not
				 * compatible.
				 */
				goto dma_vec_done;
			}
			last_phys = dma_phys + dma_count;

			/* Update counters. */
			rq->count += dma_count;
			addr += dma_count;
			size -= dma_count;

			if (rq->count == nbytes) goto dma_vec_done;
			if (rq->count == sp->count_max) goto dma_vec_done;
		}
	}
  dma_vec_done:

	/* Round rq->count down to a block sized multiple, this may require
	 * the removal of a transfer entry on a page break.
	 */
	if ((chunk = rq->count % sp->block_size) > 0) {
		/* Don't do these bytes yet. */
		rq->count -= chunk;
		if ((last_count -= chunk) > 0) {
			rq->dmaptr[-1].datalen = last_count;
		} else {
			if (--rq->dmaptr == rq->dmalist) return(EIO);
		}
	}

	show_req();

	switch (sp->devtype) {
	case SCSI_DEVCDROM:
	case SCSI_DEVWORM:
	case SCSI_DEVDISK:
	case SCSI_DEVOPTICAL:
		/* A read or write SCSI command for a random access device. */
		block = div64u(rq->position, sp->block_size);

		if (block < (1L << 21)) {
			/* We can use a group 0 command for small disks. */
			group0();
			ccb_cmd0(rq).scsi_op = opcode == DEV_SCATTER
						? SCSI_WRITE : SCSI_READ;
			h2b24(ccb_cmd0(rq).lba, block);
			ccb_cmd0(rq).nblocks = rq->count / sp->block_size;
		} else {
			/* Large disks require a group 1 command. */
			group1();
			ccb_cmd1(rq).scsi_op = opcode == DEV_SCATTER
						? SCSI_WRITE1 : SCSI_READ1;
			h2b32(ccb_cmd1(rq).lba, block);
			h2b16(ccb_cmd1(rq).nblocks, rq->count / sp->block_size);
		}

		key = scsi_command(NULL, 0);

		if (key == SENSE_NO_SENSE) {
			/* fine */;
		} else
		if (key == SENSE_UNIT_ATT || key == SENSE_ABORTED_CMD) {
			/* Check condition?  Bus reset most likely. */
			/* Aborted command?  Maybe retrying will help. */
			if (--rq->retry > 0) continue;
			return(EIO);
		} else
		if (key == SENSE_RECOVERED) {
			/* Disk drive managed to recover from a read error. */
			printf("%s: soft read error at block %lu (recovered)\n",
				s_name(), b2h32(ccb_sense(rq).info));
			key = SENSE_NO_SENSE;
			break;
		} else {
			/* A fatal error occurred, bail out. */
			return(EIO);
		}
		break;

	case SCSI_DEVTAPE:
		/* A read or write SCSI command for a sequential access device.
		 */
		group0();
		ccb_cmd0(rq).scsi_op = opcode == DEV_SCATTER
						? SCSI_WRITE : SCSI_READ;
		ccb_cmd0(rq).fixed = sp->tfixed;
		h2b24(ccb_cmd0(rq).trlength, rq->count / sp->block_size);

		key = scsi_command(NULL, 0);

		if (key != SENSE_NO_SENSE) {
			/* Either at EOF or EOM, or an I/O error. */

			if (sense_eof(key) || sense_eom(key)) {
				/* Not an error, but EOF or EOM. */
				sp->at_eof = TRUE;
				sp->tstat.mt_dsreg = DS_EOF;

				/* The residual tells how much has not been
				 * read.
				 */
				rq->count -= sp->tstat.mt_resid
							* sp->block_size;

				if (sense_eof(key)) {
					/* Went over a filemark. */
					sp->tstat.mt_blkno =
					  - (int) (rq->count / sp->block_size);
					if (!sp->tfixed)
						sp->tstat.mt_blkno = -1;
					sp->tstat.mt_fileno++;
				}
			}
			if (sense_ili(key)) {
				/* Incorrect length on a variable block length
				 * tape.
				 */

				if (sp->tstat.mt_resid <= 0) {
					/* Large block could not be read. */
					return(EIO);
				}
				/* Small block read, this is ok. */
				rq->count -= sp->tstat.mt_resid;
				sp->tstat.mt_dsreg = DS_OK;
			}
			if (key == SENSE_RECOVERED) {
				/* Tape drive managed to recover from an error.
				 */
				printf("%s: soft %s error (recovered)\n",
					s_name(), opcode == DEV_GATHER
							? "read" : "write");
				key = SENSE_NO_SENSE;
				sp->tstat.mt_dsreg = DS_OK;
			}
			if (sp->tstat.mt_dsreg == DS_ERR) {
				/* Error was fatal. */
				return(EIO);
			}
		} else {
			sp->tstat.mt_dsreg = DS_OK;
		}
		if (!sp->tfixed) {
			/* Variable block length tape reads record by record.
			 */
			sp->tstat.mt_blkno++;
		} else {
			/* Fixed length tape, multiple blocks transferred. */
			sp->tstat.mt_blkno += rq->count / sp->block_size;
		}
		sp->need_eof = (opcode == DEV_SCATTER);
		break;
	}

	/* Book the bytes successfully transferred. */
	position = add64u(position, rq->count);
	for (;;) {
		if (rq->count < iov->iov_size) {
			/* Not done with this one yet. */
			iov->iov_addr += rq->count;
			iov->iov_size -= rq->count;
			break;
		}
		rq->count -= iov->iov_size;
		iov->iov_addr += iov->iov_size;
		iov->iov_size = 0;
		if (rq->count == 0) {
			/* The rest is optional, so we return to give FS a
			 * chance to think it over.
			 */
			return(OK);
		}
		iov++;
		nr_req--;
	}
  }
  return(OK);
}


/*===========================================================================*
 *				s_rdcdrom				     *
 *===========================================================================*/
PRIVATE int s_rdcdrom(proc_nr, opcode, position, iov, nr_req)
int proc_nr;			/* process doing the request */
int opcode;			/* DEV_GATHER or DEV_SCATTER */
u64_t position;			/* offset on device to read or write */
iovec_t *iov;			/* pointer to read or write request vector */
unsigned nr_req;		/* length of request vector */
{
/* CD-ROM's have a basic block size of 2k.  We could try to set a smaller
 * virtual block size, but many don't support it.  So we use this function.
 */
  struct scsi *sp = s_sp;
  int key;
  unsigned nbytes, offset, count;
  unsigned long block, curblock;
  phys_bytes user_base = proc_vir2phys(proc_addr(proc_nr), 0);

  /* Only do reads. */
  if (opcode != DEV_GATHER) return(EINVAL);

  rq->retry = 2;
  curblock = -1;

  while (nr_req > 0) {
	/* Number of bytes to transfer. */
	nbytes = iov->iov_size;

	/* Which position on disk and how close to EOF? */
	if (cmp64(position, s_dv->dv_size) >= 0) return(OK);  /* EOF */
	if (cmp64(add64u(position, nbytes), s_dv->dv_size) > 0)
		nbytes = diff64(s_dv->dv_size, position);
	rq->position = add64(s_dv->dv_base, position);

	block = div64u(rq->position, sp->block_size);
	if (curblock != block) {
		/* Read a block that contains (some of) the bytes wanted. */
		group1();
		ccb_cmd1(rq).scsi_op = SCSI_READ1;
		h2b32(ccb_cmd1(rq).lba, block);
		h2b16(ccb_cmd1(rq).nblocks, 1);
		key = scsi_command(tmp_buf, sp->block_size);

		if (key != SENSE_NO_SENSE) {
			if (key == SENSE_UNIT_ATT && -- rq->retry > 0)
				continue;	/* retry */
			return(EIO);
		}
		curblock = block;
	}

	/* Copy the bytes wanted to user space. */
	offset = rem64u(rq->position, sp->block_size);
	count = sp->block_size - offset;
	if (count > nbytes) count = nbytes;
	phys_copy(tmp_phys + offset, user_base + iov->iov_addr,
							(phys_bytes) count);
	position = add64u(position, count);
	iov->iov_addr += count;
	if ((iov->iov_size -= count) == 0) { iov++; nr_req--; }
  }
  return(OK);
}


/*===========================================================================*
 *				s_do_close				     *
 *===========================================================================*/
PRIVATE int s_do_close(dp, m_ptr)
struct driver *dp;
message *m_ptr;
{
  struct scsi *sp;

  if (s_prepare(m_ptr->DEVICE) == NIL_DEV) return(ENXIO);
  sp = s_sp;

  /* Unlock the medium on final close. */
  if (--sp->open_ct == 0)
	(void) scsi_simple(SCSI_REMOVAL, 0);

  /* Disks and such don't make trouble. */
  if (sp->devtype != SCSI_DEVTAPE) return(OK);

  sp->at_eof = FALSE;

  /* Write filemark if writes have been done. */
  if (sp->need_eof && sp->tstat.mt_dsreg != DS_ERR) {
	if (scsi_simple(SCSI_WREOF, 1) != SENSE_NO_SENSE) {
		printf("%s: failed to add filemark\n", s_name());
	} else {
		sp->tstat.mt_dsreg = DS_OK;
		sp->tstat.mt_blkno = 0;
		sp->tstat.mt_fileno++;
	}
  }

  /* Rewind if rewind device. */
  if (s_type == TYPE_RST) {
	if (scsi_simple(SCSI_REWIND, 1) != SENSE_NO_SENSE) {
		printf("%s: failed to rewind\n", s_name());
	} else {
		sp->tstat.mt_dsreg = DS_OK;
		sp->tstat.mt_blkno = 0;
		sp->tstat.mt_fileno = 0;
	}
  }
  return(OK);
}


/*===========================================================================*
 *				s_do_ioctl				     *
 *===========================================================================*/
PRIVATE int s_do_ioctl(dp, m_ptr)
struct driver *dp;
message *m_ptr;
{
  struct scsi *sp;

  if (s_prepare(m_ptr->DEVICE) == NIL_DEV) return(ENXIO);
  sp = s_sp;

  /* Ioctls are device specific. */
  switch (sp->devtype) {
  case SCSI_DEVDISK:
  case SCSI_DEVWORM:
  case SCSI_DEVCDROM:
  case SCSI_DEVOPTICAL:
	if (m_ptr->REQUEST == DIOCEJECT) {
		/* Eject disk. */
		if (sp->open_ct > 1) return(EBUSY);

		/* Send a start/stop command with code 2: stop and eject. */
		if (scsi_simple(SCSI_STRTSTP, 2) != SENSE_NO_SENSE)
			return(EIO);
		return(OK);
	}
	/* Call the common code for disks and disk like devices. */
	return(do_diocntl(dp, m_ptr));

  default:
	return(ENOTTY);

  case SCSI_DEVTAPE:
	break;
  }
  /* Further ioctls are for tapes. */

  if (m_ptr->REQUEST == MTIOCTOP) {
	struct mtop op;
	phys_bytes op_phys;
	long delta;
	int key;
	byte *buf = tmp_buf;

	/* Basic tape commands: rewind, space, write eof marks, ... */
	op_phys = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS, sizeof(op));
	if (op_phys == 0) return(EINVAL);
	phys_copy(op_phys, vir2phys(&op), (phys_bytes) sizeof(op));

	switch(op.mt_op) {
	case MTREW:
	case MTOFFL:
	case MTRETEN:
	case MTFSF:
	case MTFSR:
	case MTBSF:
	case MTBSR:
	case MTEOM:
		/* Write an EOF mark before spacing. */
		if (sp->need_eof && sp->tstat.mt_dsreg != DS_ERR) {
			if (scsi_simple(SCSI_WREOF, 1) != SENSE_NO_SENSE)
				return(EIO);
			sp->tstat.mt_blkno = 0;
			sp->tstat.mt_fileno++;
			sp->need_eof = FALSE;
		}
		sp->at_eof = FALSE;
	}

	switch(op.mt_op) {
	case MTREW:
	case MTOFFL:
	case MTRETEN:
	case MTERASE:
		/* Rewind, Offline, Retension, Erase. */
		switch(op.mt_op) {
		case MTOFFL:
			if (scsi_simple(SCSI_LOADUNLD, 0) != SENSE_NO_SENSE)
				return(EIO);
			sp->state &= ~S_READY;
			break;
		case MTRETEN:
			if (scsi_simple(SCSI_LOADUNLD, 3) != SENSE_NO_SENSE)
				return(EIO);
			break;
		case MTERASE:
			if (scsi_simple(SCSI_REWIND, 0) != SENSE_NO_SENSE)
				return(EIO);
			if (scsi_simple(SCSI_ERASE, 1) != SENSE_NO_SENSE)
				return(EIO);
			/* Rewind once more. */
			/*FALL THROUGH*/
		case MTREW:
			if (scsi_simple(SCSI_REWIND, 0) != SENSE_NO_SENSE)
				return(EIO);
		}
		sp->tstat.mt_dsreg = DS_OK;
		sp->tstat.mt_blkno = 0;
		sp->tstat.mt_fileno = 0;
		break;
	case MTFSF:
	case MTFSR:
	case MTBSF:
	case MTBSR:
		if (sp->tstat.mt_dsreg == DS_ERR) return(EIO);
		group0();
		ccb_cmd0(rq).scsi_op = SCSI_SPACE;
		delta = op.mt_count;
		if (op.mt_op == MTBSR) delta = -delta;
		if (op.mt_op == MTBSF) delta = -delta - 1;
		h2b24(ccb_cmd0(rq).trlength, delta);
		ccb_cmd0(rq).fixed =
				op.mt_op == MTFSR || op.mt_op == MTBSR ? 0 : 1;
		if ((key = scsi_command(NULL, 0)) != SENSE_NO_SENSE) {
			if (sense_key(key) != SENSE_NO_SENSE) return(EIO);

			if (sense_eom(key)) {
				/* Banging into end of tape. */
				if (op.mt_op == MTBSF || op.mt_op == MTBSR) {
					/* Backspacing to start of tape. */
					sp->tstat.mt_dsreg = DS_EOF;
					sp->tstat.mt_blkno = 0;
					sp->tstat.mt_fileno = 0;
				} else {
					/* Not forwards please! */
					return(EIO);
				}
			}
			if (sense_eof(key)) {
				/* Reaching a filemark. */
				sp->tstat.mt_dsreg = DS_EOF;
				sp->at_eof = TRUE;
				if (op.mt_op == MTFSR) {
					/* Forwards. */
					sp->tstat.mt_blkno = 0;
					sp->tstat.mt_fileno++;
				} else {
					/* Backwards (bad idea!) */
					sp->tstat.mt_blkno = -1;
					sp->tstat.mt_fileno--;
				}
			}
		} else {
			if (op.mt_op == MTFSR || op.mt_op == MTBSR) {
				sp->tstat.mt_blkno += delta;
			} else {
				sp->tstat.mt_blkno = 0;
				sp->tstat.mt_fileno += delta;
			}
			if (op.mt_op == MTBSF) {
				/* n+1 backwards, and 1 forward. */
				group0();
				ccb_cmd0(rq).scsi_op = SCSI_SPACE;
				h2b24(ccb_cmd0(rq).trlength, 1L);
				ccb_cmd0(rq).fixed = 1;
				if (scsi_command(NULL, 0) != SENSE_NO_SENSE)
					return(EIO);
				sp->tstat.mt_fileno++;
			}
			sp->tstat.mt_dsreg = DS_OK;
		}
		break;
	case MTWEOF:
		/* Write EOF marks. */
		if (sp->tstat.mt_dsreg == DS_ERR) return(EIO);
		if (op.mt_count < 0) return(EIO);
		if (op.mt_count == 0) return(OK);
		group0();
		ccb_cmd0(rq).scsi_op = SCSI_WREOF;
		h2b24(ccb_cmd0(rq).trlength, op.mt_count);
		if (scsi_command(NULL, 0) != SENSE_NO_SENSE) return(EIO);
		sp->tstat.mt_dsreg = DS_OK;
		sp->tstat.mt_blkno = 0;
		sp->tstat.mt_fileno += op.mt_count;
		sp->need_eof = FALSE;
		break;
	case MTEOM:
		/* Forward space to end of media. */
		if (sp->tstat.mt_dsreg == DS_ERR) return(EIO);
		do {
			group0();
			ccb_cmd0(rq).scsi_op = SCSI_SPACE;
			h2b24(ccb_cmd0(rq).trlength, 0x7FFFFF);
			ccb_cmd0(rq).fixed = 1;
			key = scsi_command(NULL, 0);
			sp->tstat.mt_blkno = 0;
			sp->tstat.mt_fileno += 0x7FFFFF;
			if (key != SENSE_NO_SENSE) {
				if (key != SENSE_BLANK_CHECK) return(EIO);
				sp->tstat.mt_fileno -= sp->tstat.mt_resid;
			}
		} while (key == SENSE_NO_SENSE);
		sp->tstat.mt_dsreg = DS_OK;
		break;
	case MTBLKZ:
	case MTMODE:
		/* Select tape block size or tape density. */

		/* Rewind tape. */
		if (scsi_simple(SCSI_REWIND, 0) != SENSE_NO_SENSE)
			return(EIO);

		sp->tstat.mt_dsreg = DS_OK;
		sp->tstat.mt_blkno = 0;
		sp->tstat.mt_fileno = 0;

		if (op.mt_op == MTBLKZ && op.mt_count == 0) {
			/* Request for variable block size mode. */
			sp->tfixed = FALSE;
			sp->block_size = 1;
		} else {
			/* First a modesense to get the current values. */
			if (scsi_simple(SCSI_MDSENSE, 255) != SENSE_NO_SENSE)
				return(EIO);

			/* Must at least have one block descriptor. */
			if (buf[3] < 8) return(EIO);
			buf[0] = 0;
			buf[1] = 0;
			/* buf[2]: buffered mode & speed */
			buf[3] = 8;
			if (op.mt_op == MTMODE)		/* New density */
				buf[4 + 0] = op.mt_count;
			/* buf[4 + 1]: number of blocks */
			buf[4 + 4] = 0;
			if (op.mt_op == MTBLKZ)		/* New block size */
				h2b24(buf + 4 + 5, (long) op.mt_count);

			/* Set the new density/blocksize. */
			if (scsi_simple(SCSI_MDSELECT, 4+8) != SENSE_NO_SENSE)
				return(EIO);
			if (op.mt_op == MTBLKZ) {
				sp->tfixed = TRUE;
				sp->block_size= op.mt_count;
			}
		}
		sp->state &= ~S_READY;
		if (scsi_probe() != OK) return(EIO);
		break;
	default:
		/* Not implemented. */
		return(ENOTTY);
	}
  } else
  if (m_ptr->REQUEST == MTIOCGET) {
	/* Request tape status. */
	phys_bytes get_phys;

	get_phys = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS,
							sizeof(sp->tstat));
	if (get_phys == 0) return(EINVAL);

	if (sp->tstat.mt_dsreg == DS_OK) {
		/* Old error data is never cleared (until now). */
		sp->tstat.mt_erreg = 0;
		sp->tstat.mt_resid = 0;
	}
	phys_copy(vir2phys(&sp->tstat), get_phys,
					(phys_bytes) sizeof(sp->tstat));
  } else {
	/* Not implemented. */
	return(ENOTTY);
  }
  return(OK);
}


/*===========================================================================*
 *				scsi_simple				     *
 *===========================================================================*/
PRIVATE int scsi_simple(opcode, count)
int opcode;				/* SCSI opcode */
int count;				/* count or flag */
{
/* The average group 0 SCSI command with just a simple flag or count. */

  int len = 0;	/* Sometimes a buffer is used. */
  byte *buf = tmp_buf;

  group0();
  ccb_cmd0(rq).scsi_op = opcode;

  /* Fill in the count argument at the proper place. */
  switch (opcode) {
  case SCSI_REQSENSE:
	buf = (byte *) &ccb_sense(rq);
	/* fall through */
  case SCSI_INQUIRY:
  case SCSI_MDSENSE:
  case SCSI_MDSELECT:
	ccb_cmd0(rq).nblocks = count;
	len = count;
	break;

  case SCSI_STRTSTP:
    /* SCSI_LOADUNLD: (synonym) */
  case SCSI_REMOVAL:
	ccb_cmd0(rq).nblocks = count;
	break;

  case SCSI_RDLIMITS:
	len = count;
	break;

  case SCSI_WREOF:
	h2b24(ccb_cmd0(rq).trlength, (long) count);
	break;

  case SCSI_REWIND:
  case SCSI_ERASE:
	ccb_cmd0(rq).fixed = count;
	break;
  }
  return(scsi_command(buf, len));
}


/*===========================================================================*
 *				group0					     *
 *===========================================================================*/
PRIVATE void group0()
{
  /* Prepare the ccb for a group 0 SCSI command. */

  rq->ccb.cmdlen = sizeof(cdb0_t);

  /* Clear cdb to zeros the ugly way. */
  * (u32_t *) (rq->ccb.cmd + 0) = 0;
  * (u16_t *) (rq->ccb.cmd + 4) = 0;
}


/*===========================================================================*
 *				group1					     *
 *===========================================================================*/
PRIVATE void group1()
{
  rq->ccb.cmdlen = sizeof(cdb1_t);
  * (u32_t *) (rq->ccb.cmd + 0) = 0;
  * (u32_t *) (rq->ccb.cmd + 4) = 0;
  * (u16_t *) (rq->ccb.cmd + 8) = 0;
}


/*===========================================================================*
 *				scsi_command				     *
 *===========================================================================*/
PRIVATE int scsi_command(data, len)
byte *data;
int len;
{
/* Execute a SCSI command and return the results.  Unlike most other routines,
 * this routine returns the sense key of a SCSI command instead of OK or EIO.
 * We assign SENSE HARDWARE to happenings we can't deal with.
 */
  int hard_err, key, r;

  dump_scsi_cmd();
  out_byte(SPC_INTS, 0xFF);
  if (spc_irq >= 0)
	enable_irq(spc_irq);
  rq->ccb.cmd[1] |= s_sp->lun << 5;

check_condition:
  /* Selection */
  if (spc_phase() == SPC_FREE) {
	dump_regs("selection");
	out_byte(SPC_PCTL, 0);
	out_byte(SPC_TEMP, (1 << HOST_ID) | (1 << s_sp->targ));
	out_byte(SPC_TCH, (4000L * SCSI_TIMEOUT - 15) >> 16 & 0xFF);
	out_byte(SPC_TCM, (4000L * SCSI_TIMEOUT - 15) >> 8 & 0xFF);
	out_byte(SPC_TCL, 4);
	out_byte(SPC_CMD, SPC_SELECT);
	wait_int();
	milli_delay(SCSI_DELAY);
	if (spc_phase() == SPC_FREE) {
		printf("%s: not selected\n", s_name());
		return(SENSE_HARDWARE);
	}
  }

  /* Command phase */
  if (bus_phase() != SCSI_COMMAND) {
	printf("SCSI protocol halts in %s phase\n\n", phase[bus_phase()]);
	return(SENSE_ABORTED_CMD);
  }
  dump_regs("command phase");
  scsi_out(SCSI_COMMAND, rq->ccb.cmd, rq->ccb.cmdlen);

  /* Data phase */
  switch (rq->ccb.cmd[0]) { /* opcode */
  case SCSI_WRITE:
  case SCSI_WRITE1:
  case SCSI_READ:
  case SCSI_READ1:
	/* SCSI read/write, perform hardware transfer */
	rq->dmaptr = rq->dmalist;
	rq->remain = rq->count;
	if (rq->opcode == DEV_GATHER) {
		ready_bit = SPC_DREG_FULL;
		port_rw_byte = port_read_byte;
		out_byte(SPC_PCTL, SCSI_DATA_IN);
	} else {
		ready_bit = SPC_DREG_EMPTY;
		port_rw_byte = port_write_byte;
		out_byte(SPC_PCTL, SCSI_DATA_OUT);
	}
	out_byte(SPC_TCH, (rq->remain >> 16) & 0xFF);
	out_byte(SPC_TCM, (rq->remain >> 8) & 0xFF);
	out_byte(SPC_TCL, rq->remain & 0xFF);
	dump_regs("hardware transfer");
	if (spc_dma >= 0) {
		/* Perform DMA transfer */
		out_byte(SPC_DGC, SPC_XFER_DA);
		out_byte(SPC_CMD, SPC_TRANSFER | SPC_DMA);
		while (rq->remain > 0) {
			if ((dptr & 0xFFFF) + dlen < 0x10000L
			   && dptr + rq->remain < 0x1000000L) {
				/* No 64 KB boundary, below 16 MB */
				i8237_setup(spc_dma, rq->opcode, dptr, dlen);
				wait_dma(spc_dma);
			} else {
				/* Post 8237 era memory world */
				if (rq->opcode == DEV_SCATTER)
					phys_copy(dptr, tmp_phys, dlen);
				i8237_setup(spc_dma, rq->opcode, tmp_phys, dlen);
				wait_dma(spc_dma);
				if (rq->opcode == DEV_SCATTER)
					phys_copy(tmp_phys, dptr, dlen);
			}
			rq->remain -= dlen;
			rq->dmaptr++;
		}
	} else {
		/* perform PIO transfer */
		out_byte(SPC_DGC, SPC_XFER_EA);
		out_byte(SPC_CMD, SPC_TRANSFER | SPC_PIO);
		if (spc_irq < 0) spc_pio();
	}
	wait_int();
	milli_delay(SCSI_DELAY);
	rq->count -= rq->remain;
	hard_err = rq->remain;
	break;
  default:
	/* simple commands, perform manual transfer */
	dump_regs("data phase");
	switch (bus_phase()) {
	case SCSI_DATA_IN:
		scsi_in(SCSI_DATA_IN, data, len);
		break;
	case SCSI_DATA_OUT:
		scsi_out(SCSI_DATA_OUT, data, len);
		break;
	default:
		break;
	} /* switch bus_phase() */
	hard_err = 0;
	break;
  } /* switch opcode */

  key = r = 0;

  /* Status phase */
  dump_regs("status phase");
  scsi_in(SCSI_STATUS, (byte *) &key, 1);

  /* Message phase */
  if (bus_phase() == SCSI_MSG_IN) {
	dump_regs("message phase");
	scsi_in(SCSI_MSG_IN | (spc_irq >= 0 ? SPC_BFI_EA : SPC_BFI_DA),
		(byte *) &r, 1);
	wait_int();
  }

  if (hard_err) key = -1;		/* Hardware transfer failed */

  /* Check up the result and perform REQUEST SENSE in case of check condition */
  switch (key) {
  case GOOD:
	if (rq->ccb.cmd[0] == SCSI_REQSENSE)
		key = data[2] & 0x0F;
	else
		key = r == COMMAND_COMPLETE ? SENSE_NO_SENSE : SENSE_HARDWARE;
	break;
  case CHECK_CONDITION:
	if (rq->ccb.cmd[0] == SCSI_REQSENSE) {
		/* don't retry request sense */
		key = SENSE_HARDWARE;
	} else {
		/* last command resulted in check condition,
		   try request sense */
		group0();
  		rq->ccb.cmd[0] = SCSI_REQSENSE;
		len = rq->ccb.cmd[4] = sizeof(sense_t);
		data = (byte *) &ccb_sense(rq);
		goto check_condition;
		/* never reachable */
	}
	break;
  case BUSY:
	key = SENSE_ABORTED_CMD;
	break;
  default:
	key = SENSE_HARDWARE;
	break;
  }

  if (spc_irq >=0 )
	disable_irq(spc_irq);
  return(key);
}


/*===========================================================================*
 *				spc_pio					     *
 *===========================================================================*/
PRIVATE void spc_pio()
{
  /* perform PIO transfer */

  int i;

  while (rq->remain > 0) {
	for (i = dlen / s_sp->block_size; i > 0; i--) {
		while (!(in_byte(SPC_STS) & ready_bit))
			;
		(*port_rw_byte)(SPC_DREG, dptr, s_sp->block_size);
		dptr += s_sp->block_size;
	}
	rq->remain -= dlen;
	rq->dmaptr++;
  }

  if (i = spc_tc()) {
	/* Crazy fast CPU?  PIO overrun */
	if (rq->opcode == DEV_SCATTER)
		scsi_in(SCSI_DATA_IN, tmp_buf, i);
	else
		scsi_out(SCSI_DATA_OUT, tmp_buf, i);
	printf("%s: PIO transfer failed.  Try DMA transfer.\n", s_name());
	rq->remain = rq->count;
  }
  return;
}


/*===========================================================================*
 *				scsi_in					     *
 *===========================================================================*/
PRIVATE void scsi_in(phase, data, len)
int phase;
byte *data;
int len;
{
  /* receive from the target device by manual transfer */

  out_byte(SPC_PCTL, phase);
  phase &= PHASE_MASK;

  while (len-- > 0) {
	while (!(in_byte(SPC_PSNS) & SCSI_REQ))
		;
#if SCSI_DEBUG & 1
	printf("%x ", *data++ = in_byte(SPC_TEMP));
#else
	*data++ = in_byte(SPC_TEMP);
#endif
	out_byte(SPC_CMD, SPC_SET_AR);
	while (in_byte(SPC_PSNS) & SCSI_REQ)
		;
	out_byte(SPC_CMD, SPC_RESET_AR);
	if (bus_phase() != phase) break;
  }
#if SCSI_DEBUG & 1
	printf("\n");
#endif
	milli_delay(SCSI_DELAY);
  return;
}


/*===========================================================================*
 *				scsi_out				     *
 *===========================================================================*/
PRIVATE void scsi_out(phase, data, len)
int phase;
byte *data;
int len;
{
  /* send out to the target device by manual transfer */

  out_byte(SPC_PCTL, phase);
  phase &= PHASE_MASK;

  while (len-- > 0) {
	while (!(in_byte(SPC_PSNS) & SCSI_REQ))
		;
#if SCSI_DEBUG & 1
	printf("%x ", *data);
#endif
	out_byte(SPC_TEMP, *data++);
	out_byte(SPC_CMD, SPC_SET_AR);
	while (in_byte(SPC_PSNS) & SCSI_REQ)
		;
	out_byte(SPC_CMD, SPC_RESET_AR);
	if (bus_phase() != phase) break;
  }
#if SCSI_DEBUG & 1
	printf("\n");
#endif
	milli_delay(SCSI_DELAY);
  return;
}


/*===========================================================================*
 *				spc_init				     *
 *===========================================================================*/
PRIVATE void spc_init()
{
  long v;
  static char spc_env[] = "SPC", spc_fmt[] = "x:d:d";
  /* kernel env SPC=IO:IRQ:DMA */

  /* Get the configuration info from the environment. */
  v = SPC_BASEREG;
  (void) env_parse(spc_env, spc_fmt, 0, &v, 0x000L, 0x3FFL);
  spc_base = v;

  v = -1;
  (void) env_parse(spc_env, spc_fmt, 1, &v, 0L, 15L);
  spc_irq = v;

  v = -1;
  (void) env_parse(spc_env, spc_fmt, 2, &v, 0L, 3L);
  spc_dma = v;

  v = spc_irq >=0 ? SPC_ARBIT | SPC_PARITY | SPC_INT : SPC_ARBIT | SPC_PARITY;

  /* Reset controller and SCSI bus. */
  for (spc_offs = 1; spc_offs < 3; spc_offs++) {
	out_byte(SPC_CTL, SPC_RESET);
	milli_delay(SCSI_DELAY);
	if (in_byte(SPC_CTL) == SPC_RESET) break;
  }
  out_byte(SPC_BDID, HOST_ID);
  if (spc_offs > 2 || in_byte(SPC_BDID) != 1 << HOST_ID) {
	printf("SPC I/O base 0x%x probe failed\n\n", spc_base);
	spc_status = EIO;
	return;
  }

  out_byte(SPC_CMD, 0);
  out_byte(SPC_PCTL, 0);
  out_byte(SPC_TCH, 0);
  out_byte(SPC_TCM, 0);
  out_byte(SPC_TCL, 0);
  out_byte(SPC_TEMP, 0);
  out_byte(SPC_CTL, SPC_RESET | v);
  milli_delay(SCSI_DELAY);
  out_byte(SPC_CTL, v);
  milli_delay(SCSI_DELAY);
  printf("spc: I/O base 0x%x, offset %d", spc_base, spc_offs);
  if (spc_irq >= 0) {
	printf(", IRQ %d", spc_irq);
	put_irq_handler(spc_irq, s_handler);
  }
  if (spc_dma >= 0) {
	if (DMA_BUF_SIZE < MAX_BLOCK_SIZE
	   || (tmp_phys & 0xFFFF) + DMA_BUF_SIZE >= 0x10000L) {
		spc_dma = -1;
		printf("\nBuffer is not available and DMA is disabled");
	} else {
		printf(", DMA channel %d", spc_dma);
	}
  }
  printf("\n\n");

  spc_status = OK;
  return;
}


/*===========================================================================*
 *				s_handler				     *
 *===========================================================================*/
PRIVATE int s_handler(irq)
int irq;
{
  /* interrupt handler for both block wait and interrupt-driven hardware
     transfer */

  int r;

  if (r = in_byte(SPC_INTS)) {
	/* normally command complete or disconnected interrupt */
	if (r & SPC_DISCOND)
		out_byte(SPC_PCTL, SPC_BFI_DA);
	out_byte(SPC_INTS, r);
	if (r & SPC_DISCOND | SPC_CMD_CMPL)
		interrupt(spc_tasknr);
  } else {
	/* perform PIO transfer */
	spc_pio();
  }
  return 1;
}


/*===========================================================================*
 *				h2b16					     *
 *===========================================================================*/
PRIVATE void h2b16(b, h)
big16 b;
U16_t h;
{
/* Host byte order to Big Endian conversion. */
  b[0] = h >> 8;
  b[1] = h >> 0;
}


/*===========================================================================*
 *				h2b24					     *
 *===========================================================================*/
PRIVATE void h2b24(b, h)
big24 b;
u32_t h;
{
  b[0] = h >> 16;
  b[1] = h >>  8;
  b[2] = h >>  0;
}


/*===========================================================================*
 *				h2b32					     *
 *===========================================================================*/
PRIVATE void h2b32(b, h)
big32 b;
u32_t h;
{
  b[0] = h >> 24;
  b[1] = h >> 16;
  b[2] = h >>  8;
  b[3] = h >>  0;
}


/*===========================================================================*
 *				b2h16					     *
 *===========================================================================*/
PRIVATE u16_t b2h16(b)
big16 b;
{
  return  ((u16_t) b[0] << 8)
	| ((u16_t) b[1] << 0);
}


/*===========================================================================*
 *				b2h24					     *
 *===========================================================================*/
PRIVATE u32_t b2h24(b)
big24 b;
{
  return  ((u32_t) b[0] << 16)
	| ((u32_t) b[1] <<  8)
	| ((u32_t) b[2] <<  0);
}


/*===========================================================================*
 *				b2h32					     *
 *===========================================================================*/
PRIVATE u32_t b2h32(b)
big32 b;
{
  return  ((u32_t) b[0] << 24)
	| ((u32_t) b[1] << 16)
	| ((u32_t) b[2] <<  8)
	| ((u32_t) b[3] <<  0);
}


#if SCSI_DEBUG & 4
/*===========================================================================*
 *				show_iov				     *
 *===========================================================================*/
PRIVATE void show_iov(iov, nr_req)
iovec_t *iov;
unsigned nr_req;
{
  printf("iov:");
  while (nr_req > 0) {
	printf(" [%lx,%u]",
		(unsigned long) iov->iov_addr, (unsigned) iov->iov_size);
	iov++;
	nr_req--;
  }
  printf("\n");
}


/*===========================================================================*
 *				show_req				     *
 *===========================================================================*/
PRIVATE void show_req()
{
  dma_t *dmap;
  unsigned count;

  dmap = rq->dmalist;
  count = rq->count;

  printf("%lu:%u", cv64ul(rq->position), count);

  while (count > 0) {
	printf(" (%lx,%lu)", dmap->dataptr, dmap->datalen);
	count -= dmap->datalen;
	dmap++;
  }
  printf("\n");
}
#endif /* SCSI_DEBUG & 4 */


#if SCSI_DEBUG & 8
/*===========================================================================*
 *				dump_scsi_cmd				     *
 *===========================================================================*/
PRIVATE void dump_scsi_cmd()
{
  int i;

  printf("scsi cmd:");
  for (i = 0; i < rq->ccb.cmdlen; i++) printf(" %02x", rq->ccb.cmd[i]);
  printf("\n");
}
#endif /* SCSI_DEBUG & 8 */


/*============================================================================*
 *				s_geometry				      *
 *============================================================================*/
PRIVATE void s_geometry(entry)
struct partition *entry;
{
/* The geometry of a SCSI drive is a complete fake, the Adaptec onboard BIOS
 * makes the drive look like a regular drive on the outside.  A DOS program
 * takes a logical block address, computes cylinder, head and sector like the
 * BIOS int 0x13 call expects, and the Adaptec turns this back into a block
 * address again.  The only reason we care is because some idiot put cylinder,
 * head and sector numbers in the partition table, so fdisk needs to know the
 * geometry.
 */
  u64_t size = s_sp->part[0].dv_size;
  unsigned heads, sectors;

  if (cmp64ul(size, 1024L * 64 * 32 * 512) < 0) {
	/* Small drive. */
	heads = 64;
	sectors = 32;
  } else {
	/* Assume that this BIOS is configured for large drives. */
	heads = 255;
	sectors = 63;
  }
  entry->cylinders = div64u(size, SECTOR_SIZE) / (heads * sectors);
  entry->heads = heads;
  entry->sectors = sectors;
}
#endif /* !ENABLE_SPC_SCSI */

